use std::collections::HashMap;

use chrono::{DateTime, Duration, Timelike, Utc};
use serde::{Deserialize, Serialize};

use crate::{
    clock,
    forecast::{round_to_last_weather_time, Forecast, ForecastSet, Rate},
};

const DATA: &'static str = include_str!("../data.json");

#[derive(Serialize, Deserialize, Debug)]
pub struct Data {
    pub db_data: SubData,
    pub fish_entries: Vec<FishMeta>,
    #[serde(skip)]
    pub forecasts: ForecastSet,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct SubData {
    #[serde(alias = "FISH")]
    pub fish: HashMap<u32, FishEntry>,
    #[serde(alias = "ZONES")]
    pub zones: HashMap<u32, Zone>,
    #[serde(alias = "WEATHER_RATES")]
    pub weather_rates: HashMap<u32, WeatherRate>,
    #[serde(alias = "FISHING_SPOTS")]
    pub fishing_spots: HashMap<u32, FishingSpot>,
    #[serde(alias = "WEATHER_TYPES")]
    pub weather_types: HashMap<u32, WeatherType>,
}

#[derive(Serialize, Deserialize, Debug, Default)]
#[serde(rename_all(deserialize = "camelCase"))]
pub struct FishEntry {
    #[serde(alias = "_id")]
    pub id: u32,
    pub start_hour: Option<f32>,
    pub end_hour: Option<f32>,
    pub location: Option<u32>,
    pub best_catch_path: Vec<u32>,
    pub predators: Vec<Vec<u32>>,
    pub intuition_length: Option<u32>,
    pub patch: f32,
    pub folklore: Option<u32>,
    pub fish_eyes: bool,
    pub big_fish: bool,
    pub weather_set: Vec<u32>,
    pub previous_weather_set: Vec<u32>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Zone {
    pub name_en: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct FishingSpot {
    #[serde(alias = "_id")]
    pub id: u32,
    pub name_en: String,
    #[serde(alias = "territory_id")] // not sure why it's named this, but thanks serde!
    pub forecast_id: u32,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct WeatherRate {
    pub zone_id: u32,
    pub weather_rates: Vec<Vec<u32>>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct WeatherType {
    pub name_en: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Location {
    pub id: u32,
    pub name_en: String,
    pub map_coords: (f32, f32, u32),
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct FishMeta {
    pub id: u32,
    pub name_en: String,
}

pub struct CombinedFish<'a> {
    pub entry: &'a FishEntry,
    pub meta: &'a FishMeta,
    pub is_up: bool,
    pub next_uptime: DateTime<Utc>,
}

impl<'a> CombinedFish<'a> {
    /// Fills in the rest of the struct.
    pub fn find_uptime(&mut self, data: &Data) {
        self.is_up = self.is_up_now(data);
        self.next_uptime = self.get_next_uptime(data);
    }

    /// Finds out whether the fish is technically up at the given time (discounting factors like weather).
    pub fn is_in_time_range_at(&self, date: DateTime<Utc>) -> bool {
        if self.entry.start_hour.is_none() || self.entry.end_hour.is_none() {
            return false;
        }
        let et = clock::to_eorzea_time(date);
        let start_hour = self.entry.start_hour.unwrap();
        let end_hour = self.entry.end_hour.unwrap();
        let spans_midnight = start_hour > end_hour;
        let cur_hour = et.hour() as f32 + et.minute() as f32 / 60.;
        if spans_midnight {
            return cur_hour > start_hour || cur_hour < end_hour;
        } else {
            return start_hour < cur_hour && cur_hour < end_hour;
        }
    }

    /// Gets the associated forecast struct.
    pub fn get_forecast<'b>(&self, data: &'b Data) -> Option<&'b Forecast> {
        if self.entry.location.is_none() {
            return None;
        }
        // Find the associated fishing spot.
        let spot = data
            .db_data
            .fishing_spots
            .get(&self.entry.location.unwrap())
            .unwrap();
        data.forecasts.get(&spot.forecast_id)
    }

    /// Finds out whether the fish is in favorable weather conditions at a specific Eorzean time.
    pub fn is_in_correct_weather_at(&self, forecast: &Forecast, date: DateTime<Utc>) -> bool {
        if self.entry.weather_set.is_empty() {
            return true;
        }
        // Check if the current weather is right!
        let cur_weather = forecast.weather_at(date);
        let is_current = self
            .entry
            .weather_set
            .iter()
            .any(|ws| ws == &cur_weather.weather_id);

        // Check if the fish depends on a previous weather, and if so, if that weather is
        // right as well.
        let is_past = if self.entry.previous_weather_set.is_empty() {
            true
        } else {
            let prev_date = date - Duration::hours(8);
            let prev_weather = forecast.weather_at(prev_date);
            self.entry
                .previous_weather_set
                .iter()
                .any(|ws| ws == &prev_weather.weather_id)
        };

        is_current && is_past
    }

    /// Finds out if the fish is up now, includes weather and time bounds.
    pub fn is_up_now(&self, data: &Data) -> bool {
        self.is_up_at_n(data, 0)
    }

    /// Finds out if the fish is up at a specific time, `n` Eorzean hours away.
    pub fn is_up_at_n(&self, data: &Data, n: i64) -> bool {
        let forecast = self.get_forecast(data);
        let date = clock::get_current_eorzea_date() + Duration::hours(n);
        self.is_in_time_range_at(date)
            && forecast.is_none_or(|forecast| self.is_in_correct_weather_at(forecast, date))
    }

    /// Get the next uptime for a fish that is only restricted by time.
    pub fn get_next_time_uptime(&self) -> DateTime<Utc> {
        let ez_date = clock::get_current_eorzea_date();
        // Check if we have a minute part to our start time, which can happen.
        let adjusted_date = clock::set_hm_from_float(&ez_date, self.entry.start_hour.unwrap());
        clock::to_earth_time(adjusted_date)
    }

    /// Get the next times when the weather is favorable for a fish.
    pub fn get_next_weather_uptimes(&self, data: &Data, cycles: u32) -> Vec<DateTime<Utc>> {
        let forecast = self.get_forecast(data).unwrap();
        // Skip ahead by 8-hour intervals, to quickly find the next favorable weather.
        let mut date = round_to_last_weather_time(clock::get_current_eorzea_date());
        // Start with the next weather cycle.
        date = date + Duration::hours(8);
        let mut results = Vec::new();
        for _i in 1..cycles {
            while !self.is_in_correct_weather_at(forecast, date) {
                date = date + Duration::hours(8);
            }
            results.push(date.clone());
            date = date + Duration::hours(8);
        }
        results
    }

    // Go until you find the next uptime slot, and return its date.
    pub fn get_next_uptime(&self, data: &Data) -> DateTime<Utc> {
        // Is this a time-only fish? If so, just return the next time.
        if self.entry.weather_set.is_empty() && self.entry.start_hour.is_some() {
            return clock::to_earth_time(self.get_next_time_uptime());
        }

        // If not, check if there's a weather-only fish
        let next_weather_uptimes = self.get_next_weather_uptimes(data, 50);
        if !self.entry.weather_set.is_empty() && self.entry.start_hour.is_none() {
            return clock::to_earth_time(next_weather_uptimes.first().unwrap().clone());
        }

        // Now it gets a little bit tricky: We need to find a weather slot that also includes the start point for the
        // time window.
        for date in next_weather_uptimes {
            // Does the start time lie within this window?
            let uptime_end = date + Duration::hours(8);
            let start_hour = self.entry.start_hour.unwrap();
            if date.hour() as f32 <= start_hour && start_hour <= uptime_end.hour() as f32 {
                return clock::to_earth_time(clock::set_hm_from_float(&date, start_hour));
            }
        }

        // FIXME: This should never be called?
        Utc::now()
    }
}

impl Data {
    pub fn new() -> Self {
        let json = &mut serde_json::Deserializer::from_str(DATA);
        let mut data: Self = serde_path_to_error::deserialize(json).unwrap();
        data.forecasts = data
            .db_data
            .weather_rates
            .iter()
            .map(|(id, wr)| {
                (
                    id.clone(),
                    Forecast {
                        zone_id: wr.zone_id,
                        rates: wr
                            .weather_rates
                            .iter()
                            .map(|rate| Rate {
                                weather_id: rate[0],
                                rate: rate[1],
                            })
                            .collect(),
                    },
                )
            })
            .collect();
        data
    }

    pub fn fish_with_meta(&self) -> HashMap<u32, CombinedFish> {
        self.db_data
            .fish
            .iter()
            .filter_map(|(k, v)| {
                let corresponding_meta = self.fish_entries.iter().find(|m| &m.id == k);
                corresponding_meta.map(|m| {
                    let mut cfish = CombinedFish {
                        entry: v,
                        meta: m,
                        is_up: false,            // fake default values for now
                        next_uptime: Utc::now(), // dito
                    };
                    cfish.find_uptime(self);
                    (k.clone(), cfish)
                })
            })
            .collect()
    }
}

pub fn get_weather_name(data: &Data, id: u32) -> &str {
    &data.db_data.weather_types.get(&id).unwrap().name_en
}

pub fn get_zone_name(data: &Data, id: u32) -> &str {
    &data.db_data.zones.get(&id).unwrap().name_en
}
